#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <bstring.h>
#include <stdlib.h>
#include <stdio.h>

#include "GISproto.h"

#define MASTER 10
#define SLAVE  11
#define MAXBUF 512
#define CLIENT_MAX 10
#define LOG_FILE "/tmp/taz.log"

static FILE* log;
static char buffer[MAXBUF];
struct
{
  int type;
  int len;
} static header;

void
reconMode( int fd, int mode )
{
  readAll( fd, &header, sizeof( header ) );
  readAll( fd, buffer, header.len );
  header.len = 2;
  buffer[1] = mode;
  fprintf( log, "\nsock %d in %s mode\n", fd, mode == MASTER ? "Master" : "Slave" );
  write( fd, &header, sizeof( header ) );
  write( fd, buffer, header.len );
}

void
sigpipe_handler( int )
{
  fprintf( log, "\nTaz: SIGPIPE\n" );
}

void
sigterm_handler( int )
{
  fprintf( stderr, "\nTaz: safe termination\n" );
  clearerr( log );
  fclose( log );
  exit( 0 );
}

void
sigusr1_handler( int )
{
  fprintf( stderr, "\nTaz: stream flushing\n" );
  clearerr( log );
  fflush( log );
}


void
main( int argc, char* argv[] )
{
  int comm_sock;
  int sockfd[CLIENT_MAX];
  int client_num;
  int port;
  int i, j, master, counter;

  fd_set fdset, fdset_backup;
  struct sockaddr_in host_addr;
  

  if( argc < 2 )
    {
      fprintf( stderr, "USAGE: %s port_number\n", argv[0] );
      exit(1);
    }

  //  fprintf( stderr, "No Fear, just only Taz!\n" );
  if( ( log = fopen( LOG_FILE, "a" ) ) == NULL )
    {
      perror( LOG_FILE );
      log = stderr;
    }
  
  fprintf( log, "No Fear, just only Taz!\n" );

  signal( SIGTERM, sigterm_handler );
  signal( SIGINT, sigterm_handler );
  signal( SIGUSR1, sigusr1_handler );
  signal( SIGPIPE, sigpipe_handler );
  
  for( i=0; i<CLIENT_MAX; i++ )
    sockfd[i]=0;

  counter = client_num = 0;
  master = -1;
  port = atoi ( argv[1] );
  comm_sock = socket( PF_INET, SOCK_STREAM, 0 );
  host_addr.sin_family = AF_INET;
  host_addr.sin_port = htons( port );
  host_addr.sin_addr.s_addr = htonl( INADDR_ANY );

  if( bind( comm_sock, ( struct sockaddr* )&host_addr, sizeof( host_addr ) ) < 0 )
    {
      perror( "bind" );
      exit( 1 );
    }

  if( listen( comm_sock, 1 ) < 0 )
    {
      perror( "listen" );
      exit(1);
    }

  FD_ZERO( &fdset_backup );
  FD_SET( 0, &fdset_backup );
  FD_SET( comm_sock, &fdset_backup );
  while( 1 )
    {
      fdset = fdset_backup;

      if( select( FD_SETSIZE, &fdset, NULL, NULL, NULL ) < 0 )
	{
	  perror( "select" );
	  continue;
	}

      clearerr( log );

      if( FD_ISSET( 0, &fdset ) )
	{
	  scanf( "%s", &buffer[1] );
	  buffer[0] = NOBJ;
	  header.type = 31; //DATA;
	  header.len = strlen( &buffer[1] ) + 2;
	  fprintf( log, "<-stdin, no %d:len %d, header: %d\n", ++counter, header.len, 
		   buffer[0] );
	  for( j=0; j<client_num; j++ )
	    {
	      fprintf( log, "->sock %d\n",  sockfd[j] );
	      write( sockfd[j], &header, sizeof( header ) );
	      write( sockfd[j], buffer, header.len );
	    }
	}
      
      if( FD_ISSET( comm_sock, &fdset ) )
	{
	  if( ( sockfd[client_num] = accept( comm_sock, NULL, NULL ) ) < 0 )
	    {
	      perror( "accept" );
	      continue;
	    }

	  FD_SET( sockfd[client_num], &fdset_backup );
	  if( master == -1 )
	    {
	      master = sockfd[client_num];
	      reconMode( sockfd[client_num], MASTER );
	    }
	  else
	    reconMode( sockfd[client_num], SLAVE );
	    
	  client_num++;
	  fprintf( log, "%d client(s)\n",  client_num );
	}

      for( i=0; i<client_num; i++ )
	if( FD_ISSET( sockfd[i], &fdset ) )
	  {
	    if( readAll( sockfd[i], &header, sizeof( header ) ) != 1 )
	      {
		close( sockfd[i] );
		FD_CLR( sockfd[i], &fdset_backup );

		fprintf( log, "\nsock %d out\n",  sockfd[i] );

		if( sockfd[i] == master )
		  master = -1;

		client_num--;
		for( j=0; j<client_num; j++ )
		  if( ( j >= i ) )
		    sockfd[j] = sockfd[j+1];

		fprintf( log, "%d client(s)\n",  client_num );
		continue;
	      }
	    
	    readAll( sockfd[i], buffer, header.len );
	    fprintf( log, "<-%d, no %d:len %d, header: %d\n", sockfd[i],
		     ++counter, header.len, buffer[0] );
	    for( j=0; j<client_num; j++ )
	      if( j != i )
		{
		  fprintf( log, "->sock %d\n",  sockfd[j] );
		  write( sockfd[j], &header, sizeof( header ) );
		  write( sockfd[j], buffer, header.len );
		}
	  }
      if( ferror( log ) )
	perror( LOG_FILE );
    }
}
